"""House of the bi-directional Synchronizer class."""
from typing import Any, Callable, Literal, Optional, Tuple, TypeVar, cast

from bidict import MutableBidict  # type: ignore

# until you fix the linting issue with bubop - run pyright.
from bubop import logger  # type: ignore

from item_synchronizer.helpers import (
    SideChanges,
    TypeStats,
    delete_n_pop,
    item_getter_handle_exc,
)
from item_synchronizer.resolution_strategy import ResolutionResult, ResolutionStrategy
from item_synchronizer.types import (
    ID,
    ConverterFn,
    DeleterFn,
    InserterFn,
    Item,
    ItemGetterFn,
    UpdaterFn,
)

_FuncT = TypeVar("_FuncT", bound=Callable[..., Any])


class Synchronizer:  # pylint: disable="R0903,R0902"
    """
    Synchronize items from two different sources.

    This class aims to offer an abstract and versatile way to create, update and/or delete
    items to keep two "sources" in sync.

    These "items" may range from Calendar entries, TODO task lists, or whatever else you want
    as long as the user registers the appropriate functions/methods to convert from one said
    item to another.
    """

    def __init__(
        self,
        *,
        A_to_B: MutableBidict,
        inserter_to_A: InserterFn,
        inserter_to_B: InserterFn,
        updater_to_A: UpdaterFn,
        updater_to_B: UpdaterFn,
        deleter_to_A: DeleterFn,
        deleter_to_B: DeleterFn,
        converter_to_A: ConverterFn,
        converter_to_B: ConverterFn,
        item_getter_A: ItemGetterFn,
        item_getter_B: ItemGetterFn,
        resolution_strategy: ResolutionStrategy,
        side_names: Tuple[str, str] = ("A Side", "B Side"),
    ):
        self._A_to_B: MutableBidict = A_to_B
        self._B_to_A: MutableBidict = A_to_B.inverse
        self._inserter_to_A = self._catch_exc(inserter_to_A)
        self._inserter_to_B = self._catch_exc(inserter_to_B)
        self._updater_to_A = self._catch_exc(updater_to_A)
        self._updater_to_B = self._catch_exc(updater_to_B)
        self._deleter_to_A = self._catch_exc(delete_n_pop(deleter_to_A, self._A_to_B))
        self._deleter_to_B = self._catch_exc(delete_n_pop(deleter_to_B, self._B_to_A))
        self._converter_to_A = self._catch_exc(converter_to_A)
        self._converter_to_B = self._catch_exc(converter_to_B)
        self._item_getter_A = item_getter_handle_exc(item_getter_A)
        self._item_getter_B = item_getter_handle_exc(item_getter_B)
        self._rs = resolution_strategy
        self._stats = tuple(TypeStats(name) for name in side_names)

    def _catch_exc(self, fn: _FuncT) -> _FuncT:
        """Run the decorated function and catch all exceptions."""

        def wrapper(*args, **kargs):
            try:
                return fn(*args, **kargs)
            except Exception as exc:  # pylint: disable="W0703"
                if isinstance(exc, KeyboardInterrupt):
                    raise

                if fn.__doc__ is not None:
                    desc = fn.__doc__.split("\n")[0].strip().rstrip(".")
                else:
                    desc = str(fn)
                logger.error(f"[{desc}] Operation failed.")
                logger.opt(exception=True).debug(f"[{desc}] Operation failed.")
                return None

        return cast(_FuncT, wrapper)

    def _convert_n_insert(self, id_: ID, insert_to_side: Literal["A", "B"]) -> Optional[ID]:
        if insert_to_side == "A":
            item_getter = self._item_getter_B
            converter = self._converter_to_A
            inserter = self._inserter_to_A
        elif insert_to_side == "B":
            item_getter = self._item_getter_A
            converter = self._converter_to_B
            inserter = self._inserter_to_B
        else:
            raise RuntimeError()

        item = item_getter(id_)
        if item is None:
            return None
        converted_item = converter(item)
        if converted_item is None:
            return None
        new_id: ID = inserter(converted_item)

        if insert_to_side == "A":
            self._stats[0].create_new()
        else:
            self._stats[1].create_new()

        return new_id

    def _convert_n_update_to_A(self, id_A: ID, item: Item):
        assert item is not None
        converted_item = self._converter_to_A(item)
        if converted_item is None:
            return

        self._updater_to_A(id_A, converted_item)
        self._stats[0].update()

    def _convert_n_update_to_B(self, id_B: ID, item: Item):
        assert item is not None
        converted_item = self._converter_to_B(item)
        if converted_item is None:
            return

        self._updater_to_B(id_B, converted_item)
        self._stats[1].update()

    def sync(self, changes_A: SideChanges, changes_B: SideChanges):
        """
        Main method for running a full bi-directional sync given changes from the two sides
        involved.

        This is the main method you are supposed to call after the instance initialization to
        actually synchronize the two sides.
        """
        try:
            return self._sync(changes_A=changes_A, changes_B=changes_B)
        finally:
            logger.warning(f"\n\n{self._stats[0]}\n{self._stats[1]}")

    def _sync_new_items(self, changes_A: SideChanges, changes_B: SideChanges):
        """
        Sync only the new items from each side.

        Items that are new on either side should have no problem getting added to the other
        insert_to_side.
        """
        props = (
            (self._A_to_B, changes_A.new, "B"),
            (self._B_to_A, changes_B.new, "A"),
        )
        for map_, new_changes, insert_to_side in props:
            for id_ in new_changes:
                insert_to_side = cast(Literal["A", "B"], insert_to_side)
                inserted_id = self._convert_n_insert(id_, insert_to_side)
                if inserted_id is None:
                    continue
                map_[id_] = inserted_id

    def _sync(
        self, changes_A: SideChanges, changes_B: SideChanges
    ):  # pylint: disable="R0912,R0915,R0914"
        self._sync_new_items(changes_A=changes_A, changes_B=changes_B)

        # items modified on both sides --------------------------------------------------------
        touched_from_B = changes_B.modified.union(changes_B.deleted)
        touched_from_B_in_A_map = {self._B_to_A[change]: change for change in touched_from_B}
        touched_from_A = changes_A.modified.union(changes_A.deleted)
        touched_from_A_in_B_map = {self._A_to_B[change]: change for change in touched_from_A}

        def format_conflict_id(conflict: ID) -> str:
            prefix_mod = "Modified from"
            prefix_del = "Deleted from"

            # find whether the A ID was deleted or modified
            conflict_in_A = touched_from_A_in_B_map[conflict]
            if conflict_in_A in changes_A.deleted:
                a_str = f"{prefix_del} A"
            elif conflict_in_A in changes_A.modified:
                a_str = f"{prefix_mod} A"
            else:
                logger.exception(
                    f"Programmatic Error regarding conflict ID [{conflict} / {conflict_in_A}]"
                )

            if conflict in changes_B.deleted:
                b_str = f"{prefix_del} B"
            elif conflict in changes_B.modified:
                b_str = f"{prefix_mod} B"
            else:
                logger.exception(
                    f"Programmatic Error regarding conflict ID [{conflict} / {conflict_in_A}]"
                )

            s = f"- [B] {conflict} / [A] {conflict_in_A}\n  {b_str}\n  {a_str}"  # type: ignore
            return s

        conflicts_in_B = touched_from_B.intersection(touched_from_A_in_B_map)
        if conflicts_in_B:
            s = "\n\n"
            s += "Modified items on both sides:\n\n"
            s += "\n".join([format_conflict_id(conflict) for conflict in conflicts_in_B])
            s += f"\n\nResolution strategy: {self._rs.__class__.__name__}"
            logger.opt(lazy=True).debug(s)

        for conflict_in_B in conflicts_in_B:
            conflict_in_A = self._B_to_A[conflict_in_B]

            # find the items in conflict
            item_B = self._item_getter_B(conflict_in_B)
            item_A = self._item_getter_A(conflict_in_A)
            result = self._rs.resolve(item_A=item_A, item_B=item_B)
            if result.result_id == ResolutionResult.ID.Mix:
                raise RuntimeError("Can't handle mixed results at the moment.")
            elif result.result_id == ResolutionResult.ID.A:
                if item_A is None:
                    if conflict_in_B in changes_B.deleted:
                        # item already deleted - just remove it from the mapping
                        self._B_to_A.pop(conflict_in_B)
                    else:
                        self._deleter_to_B(conflict_in_B)
                        self._stats[1].delete()
                else:
                    self._convert_n_update_to_B(conflict_in_B, item_A)
            else:
                if item_B is None:
                    if conflict_in_A in changes_A.deleted:
                        # item already deleted - just remove it from the mapping
                        self._A_to_B.pop(conflict_in_A)
                    else:
                        self._deleter_to_A(conflict_in_A)
                        self._stats[0].delete()
                else:
                    self._convert_n_update_to_A(conflict_in_A, item_B)

        not_conflicts_from_B = touched_from_B.difference(touched_from_A_in_B_map)
        not_conflicts_from_A = touched_from_A.difference(touched_from_B_in_A_map)

        # delete and update at will
        for id_B in not_conflicts_from_B:
            id_A = self._B_to_A[id_B]
            if id_B in changes_B.modified:
                item_B = self._item_getter_B(id_B)
                self._convert_n_update_to_A(id_A, item_B)
            elif id_B in changes_B.deleted:
                self._deleter_to_A(id_A)
                self._stats[0].delete()
            else:
                raise RuntimeError("Programmatic Error")
        for id_A in not_conflicts_from_A:
            id_B = self._A_to_B[id_A]
            if id_A in changes_A.modified:
                item_A = self._item_getter_A(id_A)
                self._convert_n_update_to_B(id_B, item_A)
            elif id_A in changes_A.deleted:
                self._deleter_to_B(id_B)
                self._stats[1].delete()
            else:
                raise RuntimeError("Programmatic Error")



"""Types used across this package."""
from datetime import datetime
from typing import Any, Callable

# Type of the identifier of an item in each side.
ID = str

# Type of an arbitrary item
Item = Any

# Insertion Function - when called with the contents of an item it should return
# the ID of the newly added item
InserterFn = Callable[[Item], ID]
# Update Function - update an item given by the item ID, using the (possibly partial) new
# contents specified by Item
UpdaterFn = Callable[[ID, Item], None]
# Deletion Function - Delete the item given by the specified ID
DeleterFn = Callable[[ID], None]
# Conversion Function - convert an item from the format of one source to the format of another.
ConverterFn = Callable[[Item], Item]

# Date Getter Function - Given an item should return the corresponding date of it.
DateGetterFn = Callable[[Item], datetime]

# ItemGetter Function - Given the ID of an Item of one source return the corresponding item on
# the other source. ItemGetter functions should raise "KeyError" in case the item is not
# present (e.g. was previously deleted)
ItemGetterFn = Callable[[ID], Item]



"""Home of the various resolution strategy classes."""
from abc import ABC, abstractmethod
from datetime import datetime
from enum import Enum
from typing import Callable

from item_synchronizer.types import DateGetterFn, Item


def _named(cls):
    cls.name = cls.__name__
    return cls


class ResolutionResult:
    """Result of a resolution.

    None signifise that the item is resolved to "Deleted"
    """

    class ID(Enum):
        """Represents the ID of the item chosen from the corresponding resolution."""

        A = 0
        B = 1
        Mix = 2

    def __init__(self, id: ID, item: Item):  # pylint: disable=W0622
        self._id = id
        self._item = item

    @property
    def result_id(self) -> ID:
        """Get the result of the current resolution."""
        return self._id

    @property
    def item(self) -> Item:
        """Get the item that was chosen by this resolution"""
        return self._item


class ResolutionStrategy(ABC):
    """Base class for all the resolution strategies."""

    def __init__(self, *args, **kargs):
        """
        Implementations of this should accept whatever arguments they require for the resolve
        step that follows.
        """

    def can_resolve(self) -> bool:
        """
        Implementation can override this to signify whether the current RS is ready to
        resolve items.
        """
        return True

    @abstractmethod
    def resolve(self, item_A: Item, item_B: Item) -> ResolutionResult:
        """
        Resolve a conflice between the two items and return the resolution result as well as
        the resolved version of the item.

        If an item is None, it means that it was deleted from the corresponding side.
        """
        raise NotImplementedError()

    @property
    @classmethod
    def name(cls) -> str:
        """Get the name of the derived resolution strategy."""
        raise NotImplementedError


class RecencyRS(ResolutionStrategy):
    """Base class for the resolution strategies that act based on recency of the items.

    You probably want to use one of the child classes instead of this one.
    """

    def __init__(
        self,
        date_getter_A: DateGetterFn,
        date_getter_B: DateGetterFn,
        compare_dates: Callable[[datetime, datetime], bool],
        *args,
        **kargs,
    ):
        """
        Initialize the Resolution Strategy.

        Provide at least the way of getting each one of the dates as well as the functions to
        compare them.
        """
        self._date_getter_A = date_getter_A
        self._date_getter_B = date_getter_B
        self._compare_dates = compare_dates

        super().__init__(*args, **kargs)

    def resolve(self, item_A: Item, item_B: Item) -> ResolutionResult:
        # handle None(s) ----------------------------------------------------------------------
        if item_A is None and item_B is None:
            return ResolutionResult(id=ResolutionResult.ID.A, item=item_A)
        if item_A is None:
            return ResolutionResult(id=ResolutionResult.ID.B, item=item_B)
        if item_B is None:
            return ResolutionResult(id=ResolutionResult.ID.A, item=item_A)

        # both have dates
        if self._compare_dates(self._date_getter_A(item_A), self._date_getter_B(item_B)):
            return ResolutionResult(id=ResolutionResult.ID.A, item=item_A)
        else:
            return ResolutionResult(id=ResolutionResult.ID.B, item=item_B)


@_named
class MostRecentRS(RecencyRS):
    """Return the most recent item."""

    def __init__(
        self,
        date_getter_A: DateGetterFn,
        date_getter_B: DateGetterFn,
    ):
        super().__init__(
            date_getter_A=date_getter_A,
            date_getter_B=date_getter_B,
            compare_dates=lambda date1, date2: date1 >= date2,
        )


@_named
class LeastRecentRS(RecencyRS):
    """Return the oldest item."""

    def __init__(
        self,
        date_getter_A: DateGetterFn,
        date_getter_B: DateGetterFn,
    ):
        super().__init__(
            date_getter_A=date_getter_A,
            date_getter_B=date_getter_B,
            compare_dates=lambda date1, date2: date1 <= date2,
        )


@_named
class AlwaysFirstRS(ResolutionStrategy):
    """Return the first item."""

    def resolve(self, item_A: Item, item_B: Item) -> ResolutionResult:
        return ResolutionResult(id=ResolutionResult.ID.A, item=item_A)


@_named
class AlwaysSecondRS(ResolutionStrategy):
    """Return the second item."""

    def resolve(self, item_A: Item, item_B: Item) -> ResolutionResult:
        return ResolutionResult(id=ResolutionResult.ID.B, item=item_B)


all_resolution_strategies = [AlwaysFirstRS, AlwaysSecondRS, MostRecentRS, LeastRecentRS]



"""Init."""
from .resolution_strategy import (
    AlwaysFirstRS,
    AlwaysSecondRS,
    LeastRecentRS,
    MostRecentRS,
    RecencyRS,
    ResolutionResult,
    ResolutionStrategy,
    all_resolution_strategies,
)
from .synchronizer import Synchronizer

__all__ = [
    "AlwaysFirstRS",
    "AlwaysSecondRS",
    "LeastRecentRS",
    "MostRecentRS",
    "RecencyRS",
    "ResolutionResult",
    "ResolutionStrategy",
    "Synchronizer",
    "all_resolution_strategies",
]

__version__ = "1.1.3b"



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['item_synchronizer']

package_data = \
{'': ['*']}

install_requires = \
['bidict>=0.21.4,<0.22.0', 'bubop']

setup_kwargs = {
    'name': 'item-synchronizer',
    'version': '1.1.3b0',
    'description': 'Synchronize items bi-directionally between two different sources',
    'long_description': '# Item Synchronizer\n\n<img src="https://github.com/bergercookie/item_synchronizer/raw/master/res/logo.png" alt="logo" style="zoom:50%;" />\n\n<a href="https://github.com/bergercookie/item_synchronizer/actions" alt="CI">\n<img src="https://github.com/bergercookie/item_synchronizer/actions/workflows/ci.yml/badge.svg"/></a>\n<a href="https://github.com/pre-commit/pre-commit">\n<img src="https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit&logoColor=white" alt="pre-commit"></a>\n\n<a href=\'https://coveralls.io/github/bergercookie/item_synchronizer\'>\n<img src=\'https://coveralls.io/repos/github/bergercookie/item_synchronizer/badge.svg\' alt=\'Coverage Status\' /></a>\n<a href="https://github.com/bergercookie/item_synchronizer/blob/master/LICENSE.md" alt="LICENCE">\n<img src="https://img.shields.io/github/license/bergercookie/item_synchronizer.svg" /></a>\n<a href="https://pypi.org/project/item_synchronizer/" alt="pypi">\n<img src="https://img.shields.io/pypi/pyversions/item-synchronizer.svg" /></a>\n<a href="https://badge.fury.io/py/item-synchronizer">\n<img src="https://badge.fury.io/py/item-synchronizer.svg" alt="PyPI version" height="18"></a>\n<a href="https://pepy.tech/project/item-synchronizer">\n<img alt="Downloads" src="https://pepy.tech/badge/item-synchronizer"></a>\n<a href="https://github.com/psf/black">\n<img alt="Code style: black" src="https://img.shields.io/badge/code%20style-black-000000.svg"></a>\n\n## Description\n\nSynchronize items from two different sources in a bidirectional manner.\n\nThis library aims to offer an abstract and versatile way to _create_, _update_\nand/or _delete_ items to keep two "sources" in sync.\n\nThese "items" may range from Calendar entries, TODO task lists, or whatever else\nyou want as long as the user registers the appropriate functions/methods to\nconvert from one said item to another.\n\n## Usage\n\nThe `Synchronizer` class requires the following `Callable`s to be given, for each\none of the sides. See the most up-to-date python types\n[here](https://github.com/bergercookie/item_synchronizer/blob/master/item_synchronizer/types.py)\n\n- Insertion callable: when called with the contents of an item it should create\n  and return the ID of the newly added item on the other source\n- Update callable: update an item given by the item ID, using the (possibly\n  partial) new contents specified by Item\n- Deletion callable: Delete the item given by the specified ID\n- Conversion callable: convert an item from the format of one source to the\n  format of another.\n- `Item Getter` callable: Given the ID of an Item of one source return the\n  corresponding item on the other source.\n- `A_to_B` [bidict](https://github.com/jab/bidict)\n\n  - This should be a bidict mapping IDs of A to the corresponding IDs of B and\n    vice-versa. Given this the `item_synchronizer` is responsible for keeping\n    it up to date on insertion, update and deletion events. The contents of this\n    bidict should be persistent across the various runs, thus, consider\n    pickle-ing and unpickling its contents to disk\n\nAdditionally `item_synchronizer` needs to know what items (their IDs) were\ninserted, updated and deleted during the call to its main method, `sync()`. This\nis dependent on your application at hand. You could either cache the items and\ntheir content after each run and compare them with the latest state in the\ncurrent run. Or, the API of the calendar/task manager, etc. that you are using\nmay allow you to query the items that were modified/inserted/deleted since the\nlast run.\n\n## Examples\n\nLet\'s say you want to bi-directionally synchronize your calendar events with\nyour TODO Task Manager tasks. This way, when you remove calendar event, the\ncorrespoding task will be deleted, when you add a new task in your task manager,\na new calendar event will be created and when you update the task (e.g., change\nits description or start time) the changes will reflect in the corresponding\ncalendar entry.\n\nThus, you have the following and you want to sync the `A` and `B` items.\n\n![sync0](res/drawio/sync0.drawio.svg)\n\nAs described in the previous section, `item_synchronizer` requires a set of\nfunctions which it will call when it needs to insert, update or delete an item\nfrom the corresponding side. It also requires an `A_to_B` bidict persistent\nacross its runs with the item mappings between the two sides before the very\nlatest changes. Notice that, `item_synchonizer` will be responsible for updating\nentries in this `A_to_B` bidict, you do not need to do that manually. You only\nneed to make it persist across the different runs of the `Synchronizer.sync()`\ncall (e.g., by pickle-ing and unpickle-ing it every time your application exits\nand starts again.)\n\nThus, this is the situation that `item_synchronizer` expects at its first\nrun.\n\n![sync1](res/drawio/sync1.drawio.svg)\n\nAfter the first call to `sync()` here\'s the expected results. After this call\neach event of one side will have a counterpart on the other side and that\'s also\ngoing to be reflected in the provided `A_to_B` bidict.\n\n![sync2](res/drawio/sync2.drawio.svg)\n\nSubsequent calls to `sync()` will pick up the changes and will insert any new\nitems from each side to the other side accordingly.\n\nNow let\'s say that item 2 was modified from side A, item 3 from side B and\nitem 21 was deleted from side A.\n\n![update-n-delete0](res/drawio/update-n-delete0.drawio.svg)\n\nIn the subsequent call to `sync()`, `item_synchronizer` will forward these\nchanges to the other side appropriately.\n\n![update-n-delete1](res/drawio/update-n-delete1.drawio.svg)\n\nIf there was a conflict, e.g., an item removed from one side and updated from\nthe other, then `item_synchronizer` supports a series of resolution strategies\nfor handling such conflicts.\n\n![resolution-strategy](res/drawio/resolution-strategy.drawio.svg)\n\n## Installation\n\nAdd it as a dependency to either your `requirements.txt` or to `pyproject.toml`\n\n```console\n[tool.poetry.dependencies]\n...\nitem_synchronizer = "^1.0"\n...\n```\n\nOr simply install it with `pip` if you want to use it locally:\n\n```sh\npip3 install item_synchronizer\n```\n\n## Projects using it\n\nProjects using this:\n\n- [Taskwarrior <-> Google Calendar Bidirectonal Synchronisation](https://github.com/bergercookie/taskw_gcal_sync/blob/master/taskw_gcal_sync/TWGCalAggregator.py)\n\n## Notes\n\n- Currently IDs of items on either side should be of `str` type.\n',
    'author': 'Nikos Koukis',
    'author_email': 'nickkouk@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/bergercookie/item_synchronizer',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""Helper functions and classes"""
from dataclasses import dataclass, field
from typing import MutableMapping, Set

from item_synchronizer.types import ID, DeleterFn, ItemGetterFn


@dataclass
class SideChanges:
    """Hold the items that are new, modified or deleted compared to the previous run.

    The items related to deletions have indeed already been deleted but the corresponding
    A <-> B mapping has not been updated yet. It's the responsibility of the Synchronizer class
    to do so
    """

    new: Set[ID] = field(default_factory=set)
    modified: Set[ID] = field(default_factory=set)
    deleted: Set[ID] = field(default_factory=set)

    def __str__(self) -> str:
        s = f"New Items:      {len(self.new)}\n\t"
        s += "\n\t".join(id_ for id_ in self.new)
        s = s.rstrip("\n\t")
        s += f"\nModified Items: {len(self.modified)}\n\t"
        s += "\n\t".join(id_ for id_ in self.modified)
        s = s.rstrip("\n\t")
        s += f"\nDeleted Item:   {len(self.deleted)}\n\t"
        s += "\n\t".join(id_ for id_ in self.deleted)
        s = s.rstrip("\t")
        return s


def item_getter_handle_exc(item_getter: ItemGetterFn) -> ItemGetterFn:
    """ItemGetter decorator function that handles exception when handing over the item."""

    def fn(*args, **kargs):
        try:
            return item_getter(*args, **kargs)
        except KeyError:
            return None

    return fn


def delete_n_pop(deleter: DeleterFn, map_: MutableMapping) -> DeleterFn:
    """Wrapper function for deleting and popping an item from the given map mapping."""
    delete_n_pop.__doc__ = deleter.__doc__

    def fn(id_):
        deleter(id_)
        map_.pop(id_)

    return fn


class TypeStats:
    """Container class for printing execution stats on exit - per type."""

    def __init__(self, title: str):
        self._title = title

        self._created_new = 0
        self._updated = 0
        self._deleted = 0
        self._errors = 0

        self._sep = "-" * len(self._title)

    def create_new(self):
        """Report an insertion event."""
        self._created_new += 1

    def update(self):
        """Report an update event."""
        self._updated += 1

    def delete(self):
        """Report a delete event."""
        self._deleted += 1

    def error(self):
        """Report an error during an event operation."""
        self._errors += 1

    def __str__(self) -> str:
        s = (
            f"{self._title}\n"
            f"{self._sep}\n"
            f"\t* Items created: {self._created_new}\n"
            f"\t* Items updated: {self._updated}\n"
            f"\t* Items deleted: {self._deleted}\n"
        )
        return s
